// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "stdafx.h"
#include "BlobShop.h"
#include "SceneFileReader.h"


#include "SceneManager.h"
#include "TextureManager.h"

#include "BlobTreeNodeSceneObject.h"
#include "BartheSmoothDifferenceSceneObject.h"	// for loading hack...


#define SCENEFILE_SERIALIZER_BASE_MACRO 0x5000
const Wml::SerializerRecord::RecordID SCENEFILE_SERIALIZER_BASE = SCENEFILE_SERIALIZER_BASE_MACRO;


const Wml::SerializerRecord::RecordID BLOBTREE_LIST_RECORD_ID = 1;
#define BLOBTREE_LIST_RECORD_TYPE SCENEFILE_SERIALIZER_BASE_MACRO + 1

const Wml::SerializerRecord::RecordID TEXTUREMANAGER_RECORD_ID = 2;
#define TEXTUREMANAGER_RECORD_TYPE SCENEFILE_SERIALIZER_BASE_MACRO + 2

const Wml::SerializerRecord::RecordID SCENEOPTIONS_RECORD_ID = 3;
#define SCENEOPTIONS_RECORD_TYPE SCENEFILE_SERIALIZER_BASE_MACRO + 3

void BlobTreeListRecord::AddBlobTreeRecord( Wml::SerializerRecord * pRootNodeRecord )
{
	m_vRootNodeIDs.push_back( pRootNodeRecord->GetRecordID() );
}



Wml::SerializerRecord * BlobTreeListRecord::Serialize( Wml::Serializer * pSerializer )
{
	Wml::SerializerRecord * pRecord = pSerializer->CreateRecord( BLOBTREE_LIST_RECORD_ID, BLOBTREE_LIST_RECORD_TYPE );

	pRecord->SetValue( "TreeListSize", (unsigned int)m_vRootNodeIDs.size() );
	pRecord->SetValue( "TreeList", m_vRootNodeIDs );

	return pRecord;
}


bool BlobTreeListRecord::Serialize( Wml::SerializerRecord * pRecord, Wml::Serializer * pSerializer )
{
	unsigned int nSize;
	if ( pRecord->GetValue( "TreeListSize", nSize, 0 ) ) {
		m_vRootNodeIDs.resize( nSize );
		pRecord->GetValue( "TreeList", m_vRootNodeIDs );
	}

	return true;
}


Wml::SerializerRecord * TextureManagerRecord::Serialize( Wml::Serializer * pSerializer )
{
	Wml::SerializerRecord * pRecord = pSerializer->CreateRecord( TEXTUREMANAGER_RECORD_ID, TEXTUREMANAGER_RECORD_TYPE );
	pRecord->SetValue( "TexManRecordID", m_vTextureManagerID );
	return pRecord;
}

bool TextureManagerRecord::Serialize( Wml::SerializerRecord * pRecord, Wml::Serializer * pSerializer )
{
	return pRecord->GetValue( "TexManRecordID", m_vTextureManagerID, 0 );
}



Wml::SerializerRecord * SceneOptionsRecord::Serialize( Wml::Serializer * pSerializer )
{
	Wml::SerializerRecord * pRecord = pSerializer->CreateRecord( SCENEOPTIONS_RECORD_ID, SCENEOPTIONS_RECORD_TYPE );
	pRecord->SetValue( "ViewRecordID", m_vViewRecordID );
	return pRecord;
}

bool SceneOptionsRecord::Serialize( Wml::SerializerRecord * pRecord, Wml::Serializer * pSerializer )
{
	return pRecord->GetValue( "ViewRecordID", m_vViewRecordID, 0 );
}




SceneFileReader::SceneFileReader( CSceneManager * pSceneManager, 
								  CTextureManager * pTextureManager, 
								  CViewController * pViewController )
	: m_pSceneManager(pSceneManager),
	  m_pTextureManager(pTextureManager),
	  m_pViewController(pViewController)
{
	m_pSerializer = NULL;
}

SceneFileReader::~SceneFileReader()
{
}

bool SceneFileReader::RestoreScene( Wml::Serializer * pSerializer )
{
	m_pSerializer = pSerializer;

	// restore blobtree list

	Wml::SerializerRecord * pBTListRecord = m_pSerializer->FindRecord( BLOBTREE_LIST_RECORD_ID );
	if ( pBTListRecord == NULL )
		return false;

	BlobTreeListRecord * pBTList = new BlobTreeListRecord();
	if ( ! pBTList->Serialize( pBTListRecord, m_pSerializer ) )
		return false;

	const std::vector< Wml::SerializerRecord::RecordID > & vRootNodes = pBTList->GetRootNodeIDs();
	size_t nCount = vRootNodes.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {

		Wml::SerializerRecord * pRootNodeRecord = m_pSerializer->FindRecord( vRootNodes[i] );
		if ( pRootNodeRecord ) {
			rmsui_bt::BlobTreeNodeSceneObject * pRootNode = RestoreBTNodeRecord( pRootNodeRecord, NULL );
			if ( ! pRootNode )
				return false;
		}
	}

	// Restore texture manager. If not found, assume there are no textures...
	Wml::SerializerRecord * pTexManRec = m_pSerializer->FindRecord( TEXTUREMANAGER_RECORD_ID );
	if ( pTexManRec != NULL ) {
		TextureManagerRecord tmpRec;
		if (! tmpRec.Serialize( pTexManRec, m_pSerializer ) )
			return false;

		Wml::SerializerRecord * pRealRecord = m_pSerializer->FindRecord( tmpRec.TexManRecordID() );
		if ( pRealRecord == NULL )
			return false;

		m_pTextureManager->Serialize( pRealRecord, pSerializer );
	}

	// restore scene options. If not found, the don't...
	Wml::SerializerRecord * pSceneOptionsRec = m_pSerializer->FindRecord( SCENEOPTIONS_RECORD_ID );
	if ( pSceneOptionsRec != NULL ) {
		SceneOptionsRecord tmpRec;
		if (! tmpRec.Serialize( pSceneOptionsRec, m_pSerializer ) )
			return false;

		// restore view
		Wml::SerializerRecord * pRealRecord = m_pSerializer->FindRecord( tmpRec.ViewRecordID() );
		if ( pRealRecord == NULL )
			return false;
		m_pViewController->Serialize(pRealRecord, pSerializer);
	}

	return true;
}




void SceneFileReader::RestoreBTNodeChildren( Wml::SerializerRecord * pRecord, rmsui_bt::BlobTreeNodeSceneObject * pParentNode )
{
	unsigned int nChildren;
	if ( pRecord->GetValue( "NumChildren", nChildren, 0 ) == false )
		return;

	std::vector< Wml::SerializerRecord::RecordID > vChildIDs;
	vChildIDs.resize(nChildren);

	if ( pRecord->GetValue( "ChildrenIDs", vChildIDs ) == false )
		return;

	size_t nCount = vChildIDs.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {

		Wml::SerializerRecord * pChildRecord = m_pSerializer->FindRecord( vChildIDs[i] );
		if ( pChildRecord ) {
			RestoreBTNodeRecord( pChildRecord, pParentNode );
		}
	}
}

rmsui_bt::BlobTreeNodeSceneObject * SceneFileReader::RestoreBTNodeRecord( Wml::SerializerRecord * pRecord, rmsui_bt::BlobTreeNodeSceneObject * pParent )
{
	rmsui_bt::BlobTreeNodeSceneObject * pRestoredNode = NULL;

	// create temporary thing to switch on...
	unsigned int nShiftedType = pRecord->GetRecordType() - BTNODE_SO_SERIALIZER_BASE;
	rmsui_bt::BlobTreeNodeSceneObject::Type eType = (rmsui_bt::BlobTreeNodeSceneObject::Type)nShiftedType;

	switch( eType ) {

		//case BLOBTREE_LIST_RECORD_TYPE:
		//	{
		//		BlobTreeListRecord * pBTList = new BlobTreeListRecord();
		//		pBTList->Serialize( pRecord, m_pSerializer );
		//		m_pSerializer->AddItem( pRecord->GetRecordID(), pBTList );
		//	}


		case rmsui_bt::BlobTreeNodeSceneObject::RootNode:
			{
				rmsui_bt::RootNodeSceneObject * pRootNode = m_pSceneManager->GetCurrentBlobTree()->GetRootNode();
				pRootNode->Serialize( pRecord, m_pSerializer );
				m_pSerializer->AddItem( pRecord->GetRecordID(), pRootNode );

				RestoreBTNodeChildren( pRecord, pRootNode );

				pRestoredNode = pRootNode;
				break;
			}


		case rmsui_bt::BlobTreeNodeSceneObject::SumBlend:
		case rmsui_bt::BlobTreeNodeSceneObject::MinMaxUnion:
		case rmsui_bt::BlobTreeNodeSceneObject::MinMaxDifference:
		case rmsui_bt::BlobTreeNodeSceneObject::MinMaxIntersection:
		case rmsui_bt::BlobTreeNodeSceneObject::BartheSharpIntersection:
		case rmsui_bt::BlobTreeNodeSceneObject::BartheSharpUnion:
		case rmsui_bt::BlobTreeNodeSceneObject::BartheSharpDifference:
		case rmsui_bt::BlobTreeNodeSceneObject::BartheSmoothDifference:
		case rmsui_bt::BlobTreeNodeSceneObject::RicciBlend:
			{
				bool bHack = (eType == rmsui_bt::BlobTreeNodeSceneObject::MinMaxDifference);
				if ( bHack )
					eType = rmsui_bt::BlobTreeNodeSceneObject::BartheSmoothDifference;

				rmsui_bt::BlobTreeNodeSceneObject * pNode = m_pSceneManager->RestoreSimpleNode( eType, pParent );
				pNode->Serialize( pRecord, m_pSerializer );
				m_pSerializer->AddItem( pRecord->GetRecordID(), pNode );

				if ( bHack ) {
					( (rmsui_bt::BartheSmoothDifferenceSceneObject *) pNode )->GetBartheSmoothDifferenceNode()->SetSharpness(0.9f);
				}

				RestoreBTNodeChildren( pRecord, pNode );

				pRestoredNode = pNode;
				break;
			}

		case rmsui_bt::BlobTreeNodeSceneObject::CappedLinearSweep:
		case rmsui_bt::BlobTreeNodeSceneObject::Revolution:
		case rmsui_bt::BlobTreeNodeSceneObject::Point:
		case rmsui_bt::BlobTreeNodeSceneObject::Cylinder:
		case rmsui_bt::BlobTreeNodeSceneObject::PolyLine:
		case rmsui_bt::BlobTreeNodeSceneObject::InterpolatingLinearSweep:
			{
				rmsui_bt::BlobTreeNodeSceneObject * pNode = 
					m_pSceneManager->RestoreSimpleNode( eType, pParent );
				pNode->Serialize( pRecord, m_pSerializer );
				m_pSerializer->AddItem( pRecord->GetRecordID(), pNode );

				pRestoredNode = pNode;
				break;
			}
	}

	return pRestoredNode;
}
