#include "StdAfx.h"
#include "Constraint.h"
#include "UIGlobals.h"
#include "rmsdebug.h"

#include <WmlLine3.h>
#include <WmlExtPlane3.h>
#include <WmlDistLin3Lin3.h>
#include <WmlDistVec3Lin3.h>
#include <WmlIntrLin3Tri3.h>
#include <WmlIntrLin2Lin2.h>


#include <GL/gl.h>
#include <GL/glu.h>
#include <limits>


using namespace rmsui;


//float g_fLineWidthMultiplier = 3.0f;
float g_fLineWidthMultiplier = 1.0f;


Constraint::Constraint( ConstraintTarget * pTarget ) 
{
	m_bPinned = false;
	m_eType = World;
	m_cColor = Wml::ColorRGBA::BLACK;

	m_pTarget = pTarget;
	m_pTarget->AddListener(this);
}


Constraint::~Constraint() 
{
	m_pTarget->RemoveListener(this);
}



void Constraint::SetFrame( const Wml::Frame3f & vFrame )
{
	m_vFrame = vFrame;
	UpdateRelative( m_pTarget->GetFrame() );
}


void Constraint::UpdateFrame( const Wml::Vector3f & vNewOrigin, const Wml::Vector3f & vNewAxis )
{
	Wml::Frame3f vFrame(vNewOrigin, vNewAxis);
	SetFrame(vFrame);
}


void Constraint::UpdateRelative( const Wml::Frame3f & vRelativeToFrame )
{
	m_vRelativeFrame.Origin() = m_vFrame.Origin() - vRelativeToFrame.Origin();
	Wml::Vector3f vX(m_vFrame.X());  vRelativeToFrame.ToFrameOrientation( vX );
	Wml::Vector3f vY(m_vFrame.Y());  vRelativeToFrame.ToFrameOrientation( vY );
	Wml::Vector3f vZ(m_vFrame.Z());  vRelativeToFrame.ToFrameOrientation( vZ );
	m_vRelativeFrame.SetFrame( vX, vY, vZ );
	m_vRelativeFrame.ReNormalize(2);

	// there is going to be some numerical error when calling FromRelative, might
	// as well take it now so that we don't see the popping after the first call...
	FromRelative( vRelativeToFrame );
}

void Constraint::FromRelative( const Wml::Frame3f & vFromFrame )
{
	m_vFrame.Origin() = vFromFrame.Origin() + m_vRelativeFrame.Origin();

	// screen and world frames don't rotate w/ object
	if ( AxisType() == Object || AxisType() == Surface ) {
		Wml::Vector3f vX(m_vRelativeFrame.X());  vFromFrame.FromFrameOrientation( vX );
		Wml::Vector3f vY(m_vRelativeFrame.Y());  vFromFrame.FromFrameOrientation( vY );
		Wml::Vector3f vZ(m_vRelativeFrame.Z());  vFromFrame.FromFrameOrientation( vZ );

		// try to preserve frames...
		if ( m_vFrame.X().Dot(vX) < 0.001f ) 
			vX = m_vFrame.X();
		if ( m_vFrame.Y().Dot(vY) < 0.001f ) 
			vY = m_vFrame.Y();
		if ( m_vFrame.Z().Dot(vZ) < 0.001f ) 
			vZ = m_vFrame.Z();
		
		m_vFrame.SetFrame( vX, vY, vZ );
		m_vFrame.ReNormalize(2);
	}
}





















UniformScaleConstraint::UniformScaleConstraint( ConstraintTarget * pTarget, const Wml::Vector3f & vOrigin )
	: Constraint(pTarget)
{
	m_fDrawLength = 0.5f;
	m_fDrawLineWidth = 8.0f;

	m_vOrigin = vOrigin;
	m_vPlaneNormal = Wml::Vector3f::UNIT_Z;
	m_bPlaneFrameSet = false;
}

UniformScaleConstraint::UniformScaleConstraint( ConstraintTarget * pTarget, const Wml::Frame3f & vFrame )
	: Constraint(pTarget)
{
	m_fDrawLength = 0.5f;
	m_fDrawLineWidth = 8.0f;

	m_vOrigin = vFrame.Origin();
	m_vPlaneNormal = vFrame.Z();
	m_bPlaneFrameSet = true;
	m_vPlaneFrame = vFrame;
}


void UniformScaleConstraint::UpdateGeometry( const ViewState & eViewState )
{
	// make geometry frame
	if ( m_bPlaneFrameSet ) 
		m_geomFrame = m_vPlaneFrame;
	else {
		m_geomFrame.Origin() = m_vOrigin;
	}

	// generate geometry
	float fDrawLength = m_fDrawLength * eViewState.fFixedSizeScaleFactor * UIGlobals::Get()->GetUIScaleFactor();
	fDrawLength *= 0.2f;

	m_geom.Clear();
	m_geom.Frame() = m_geomFrame;
	m_geom.CurrentColor() = Color();
	m_geom.AddCube( Color(), fDrawLength, true );

	m_vLookDir = eViewState.vLookDirection;
}



void UniformScaleConstraint::Render(const ViewState & eViewState)
{
	Wml::Vector3f vOrigin(m_vOrigin);

	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);
	glDisable(GL_LIGHTING);
	//glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

	UpdateGeometry(eViewState);

	glLineWidth(1.5f * g_fLineWidthMultiplier );

	m_geom.RenderLineColor() = 
		Pinned() ? Wml::ColorRGBA::RED : Wml::ColorRGBA::BLACK;
	m_geom.Render(true, false);

	glPopAttrib();
}


void UniformScaleConstraint::GetHitPoint( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
											float & fDistanceFromRay, float & fRayT )
{
	// find intersection between ray and plane
	Wml::ExtPlane3f plane( m_vLookDir, m_vOrigin );
	fRayT = plane.IntersectRay( ray.Origin(), ray.Direction() );

	ptHit = ray.Origin() + fRayT * ray.Direction();
	fDistanceFromRay = 0.0f;
}

//! returns true on intersection, or if point was suitably close
bool UniformScaleConstraint::Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
										float & fDistanceFromRay, float & fRayT, int & nHitTag )
{
	if ( m_geom.Intersect( ray, ptHit, nHitTag ) ) {
		GetHitPoint(ray, ptHit, fDistanceFromRay, fRayT);
		return true;
	}

	return false;
}


//! initialize constraint
void UniformScaleConstraint::InitializeConstraint( const Wml::Ray3f & ray, const Wml::Frame3f & vInitialFrame )
{
	Wml::Vector3f vHit;
	float fDistFromRay, fRayT;
	GetHitPoint(ray, vHit, fDistFromRay, fRayT);
	m_vLastHitPoint = vHit;
}


//! returns translation to apply to frame origin
bool UniformScaleConstraint::GetUniformScaleDelta( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
												   float & fScale )
{
	Wml::Vector3f vHit;
	float fDistFromRay, fRayT;
	GetHitPoint(ray, vHit, fDistFromRay, fRayT);

	Wml::Vector3f vDelta = vHit - m_vLastHitPoint;
	fScale = vDelta.X() + vDelta.Y();

	m_vLastHitPoint = vHit;
	return true;
}


bool UniformScaleConstraint::TestCross( const ViewState & eViewState, 
										  const Wml::Vector2f & v1, const Wml::Vector2f & v2, 
										  Wml::Vector3f & vHit3D )
{
	ViewStateProjector proj;
	proj.pViewState = &eViewState;
	Wml::Segment2f seg;
	seg.Origin() = v1;
	seg.Direction() = v2 - v1;
	if ( m_geom.TestLinesCross(seg, proj, vHit3D) )
		return true;
	return false;
}


void UniformScaleConstraint::OnTargetChanged(ConstraintTarget * pTarget)
{
	m_vOrigin = m_pTarget->GetFrame().Origin();
	m_vPlaneFrame.Origin() = m_vOrigin;
}

















AxisScaleConstraint::AxisScaleConstraint( ConstraintTarget * pTarget, const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection )
	: Constraint(pTarget)
{
	m_fDrawLength = 0.5f;
	m_fDrawLineWidth = 8.0f;

	m_vOrigin = vOrigin;
	m_vDirection = vDirection;

	m_vFrame = pTarget->GetFrame();
	m_vOriginRelative = m_vOrigin - m_vFrame.Origin();
	m_vDirectionRelative = m_vDirection;
	m_vFrame.ToFrameOrientation(m_vDirectionRelative);
}



void AxisScaleConstraint::UpdateGeometry( const ViewState & eViewState )
{
	// make geometry frame
	Wml::Vector3f vCameraDir = m_vOrigin - eViewState.vEyePosition;
	vCameraDir.Normalize();
	Wml::Vector3f vLineDir( m_vDirection );
	Wml::Vector3f vLookDir( vCameraDir );

	Wml::Vector3f vAxis2( vLookDir.Cross(vLineDir) );
	vAxis2.Normalize();
	Wml::Vector3f vPerp( vLineDir.Cross(vAxis2) );
	vPerp.Normalize();

	m_geomFrame.Origin() = m_vOrigin;
	m_geomFrame.SetFrame( vLineDir, vAxis2, vPerp );


	// determine visibility
	static const float fHidden = 0.003f;
	static const float fStartFade = 0.035f;
	float fDot = 1.0f - (float)abs(vLineDir.Dot(vLookDir));
	if  ( fDot < fHidden )
		m_fVisibility = 0.0f;
	else if ( fDot < fStartFade )
		m_fVisibility = (fDot - fHidden) / (fStartFade-fHidden);
	else
		m_fVisibility = 1.0f;

	// generate geometry

	float fDrawLength = m_fDrawLength * eViewState.fFixedSizeScaleFactor * UIGlobals::Get()->GetUIScaleFactor();
	float fDrawWidth = fDrawLength * 0.08f;
	float fArrowStart = fDrawLength * 0.105f; //-fDrawLength;
	float fArrowEnd = fDrawLength;

	m_geom.Clear();
	m_geom.Frame() = m_geomFrame;

	// add arrow body
	m_geom.CurrentColor() = Color();
	m_geom.AddQuad( Wml::Vector3f( fArrowEnd, fDrawWidth, 0 ),
					Wml::Vector3f( fArrowStart, fDrawWidth, 0 ),
					Wml::Vector3f( fArrowStart, -fDrawWidth, 0 ),
					Wml::Vector3f( fArrowEnd, -fDrawWidth, 0 ) );

	const float fArrowFudge = 0.001f * eViewState.fFixedSizeScaleFactor * UIGlobals::Get()->GetUIScaleFactor();
	float fArrowOffset = fArrowEnd - fArrowFudge;
	float fArrowHeight = 1.5f * fDrawLength * 0.08f;
	float fArrowWidth = 2.5f * fDrawLength * 0.08f;

	// add arrow head
	m_geom.CurrentColor() = Color();
	m_geom.AddQuad( Wml::Vector3f(fArrowOffset, fArrowHeight, 0),
					Wml::Vector3f(fArrowOffset + fArrowWidth, fArrowHeight, 0),
					Wml::Vector3f(fArrowOffset + fArrowWidth, -fArrowHeight, 0),
					Wml::Vector3f(fArrowOffset, -fArrowHeight, 0) );

	float fHaloZ = -0.005f * eViewState.fFixedSizeScaleFactor * UIGlobals::Get()->GetUIScaleFactor();
	fHaloZ = 0;

	// add border lines
	m_geom.CurrentColor() = Wml::ColorRGBA::BLACK;
	m_geom.BeginLoop();
	m_geom.AppendLoopVert( Wml::Vector3f( fArrowEnd, fDrawWidth, fHaloZ ) );
	m_geom.AppendLoopVert( Wml::Vector3f( fArrowStart, fDrawWidth, fHaloZ ) );
	m_geom.AppendLoopVert( Wml::Vector3f( fArrowStart, -fDrawWidth, fHaloZ ) );
	m_geom.AppendLoopVert( Wml::Vector3f( fArrowEnd, -fDrawWidth, fHaloZ ) );
	m_geom.AppendLoopVert( Wml::Vector3f(fArrowOffset, -fArrowHeight, fHaloZ) );
	m_geom.AppendLoopVert( Wml::Vector3f((fArrowOffset + fArrowWidth), -fArrowHeight, fHaloZ) );
	m_geom.AppendLoopVert( Wml::Vector3f((fArrowOffset + fArrowWidth), fArrowHeight, fHaloZ) );
	m_geom.AppendLoopVert( Wml::Vector3f(fArrowOffset, fArrowHeight, fHaloZ) );

	// update crossing segment
	m_vCrossingSegment.Origin() = m_geomFrame.Origin() - m_geomFrame.X();
	m_vCrossingSegment.Direction() = m_geomFrame.Origin() + m_geomFrame.X();
	m_vCrossingSegment.Direction() -= m_vCrossingSegment.Origin();

}



void AxisScaleConstraint::Render(const ViewState & eViewState)
{
	Wml::Vector3f vOrigin(m_vOrigin);

	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT | GL_LIGHTING_BIT);
	glDisable(GL_LIGHTING);
	//glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glDepthFunc(GL_LEQUAL);

	UpdateGeometry(eViewState);


	m_geom.RenderAlpha() = m_fVisibility;
	m_geom.RenderLineColor() = 
		Pinned() ? Wml::ColorRGBA::RED : Wml::ColorRGBA::BLACK;

	glLineWidth( 1.5f * g_fLineWidthMultiplier );

	m_geom.Render(true, false);

	glPopAttrib();
}


void AxisScaleConstraint::GetNearestPoint( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
										float & fDistanceFromRay, float & fRayT )
{
	// find closest point between ray and axis
	Wml::Line3f axisLine;
	axisLine.Origin() = m_vOrigin;
	axisLine.Direction() = m_vDirection;
	axisLine.Direction().Normalize();

	// compute closest point on axis to ray
	float fLineCoeff, fRayCoeff;
	float fDistSqr = Wml::SqrDistance(axisLine, ray, &fLineCoeff, &fRayCoeff);
	ptHit = Wml::Vector3f( axisLine.Origin() + axisLine.Direction() * fLineCoeff );
	fDistanceFromRay = (float)sqrt(fDistSqr);
	fRayT = fRayCoeff;
}

//! returns true on intersection, or if point was suitably close
bool AxisScaleConstraint::Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptHit, 
							      float & fDistanceFromRay, float & fRayT, int & nHitTag )
{
	if ( m_fVisibility < 0.2f )
		return false;

	if ( m_geom.Intersect( ray, ptHit, nHitTag ) ) {
		GetNearestPoint(ray, ptHit, fDistanceFromRay, fRayT);
		fDistanceFromRay = 0.0f;
		return true;
	}

	return false;
}


//! initialize constraint
void AxisScaleConstraint::InitializeConstraint( const Wml::Ray3f & ray, const Wml::Frame3f & vInitialFrame )
{
	Wml::Vector3f vHit;
	float fDistFromRay, fRayT;
	GetNearestPoint( ray, vHit, fDistFromRay, fRayT );
	m_vLastHitPoint = vHit;
}


//! returns translation to apply to frame origin
bool AxisScaleConstraint::GetScaleDelta( const Wml::Ray3f & ray, const Wml::Frame3f & vOldFrame,
										 Wml::Vector3f & vScale )
{
	Wml::Vector3f vHit;
	float fDistFromRay, fRayT;
	GetNearestPoint( ray, vHit, fDistFromRay, fRayT );

	Wml::Vector3f vDelta = vHit - m_vLastHitPoint;
	float fDir = vDelta.Dot(m_vDirection) > 0 ? 1.0f : -1.0f;
	float fScale = vDelta.Length() * fDir;

	vScale = Wml::Vector3f::ZERO;
	float fDirX = (float)abs(m_vDirection.Dot(vOldFrame.X()));
	float fDirY = (float)abs(m_vDirection.Dot(vOldFrame.Y()));
	float fDirZ = (float)abs(m_vDirection.Dot(vOldFrame.Z()));
	if ( fDirX > fDirY && fDirX > fDirZ )
		vScale.X() = fScale;
	else if ( fDirY > fDirX && fDirY > fDirZ )
		vScale.Y() = fScale;
	else
		vScale.Z() = fScale;

	m_vLastHitPoint = vHit;
	return true;
}


bool AxisScaleConstraint::TestCross( const ViewState & eViewState, 
								  const Wml::Vector2f & v1, const Wml::Vector2f & v2, 
								  Wml::Vector3f & vHit3D )
{
	ViewStateProjector proj;
	proj.pViewState = &eViewState;
	Wml::Segment2f seg;
	seg.Origin() = v1;
	seg.Direction() = v2 - v1;
	if ( m_geom.TestLinesCross(seg, proj, vHit3D) )
		return true;
	return false;
}


void AxisScaleConstraint::OnTargetChanged(ConstraintTarget * pTarget)
{
	m_vFrame = pTarget->GetFrame();

	m_vOrigin = m_vFrame.Origin() + m_vOriginRelative;
	Wml::Vector3f vOldDirection( m_vDirection );
	m_vDirection = m_vDirectionRelative;
	m_vFrame.FromFrameOrientation(m_vDirection);
}




