#include "StdAfx.h"
#include "ConstraintSet.h"
#include "UIGlobals.h"
#include "rmsdebug.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include <limits>


using namespace rmsui;


ConstraintSet::ConstraintSet()
{
}

ConstraintSet::~ConstraintSet()
{
}


void ConstraintSet::Render(const ViewState & viewState)
{
	m_viewStateCache = viewState;

	AnyConstraintSet::iterator cur(m_vConstraints.begin()), end(m_vConstraints.end());
	while ( cur != end ) 
		(*cur++).pConstraint->Render(viewState);
}

Constraint * ConstraintSet::Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptPosition, int & nHitID )
{
	float fHitThresh = 0.03f;

	Wml::Vector3f vBestHit;
	float fBestHitDistFromRay = std::numeric_limits<float>::max();
	float fBestRayT = std::numeric_limits<float>::max();
	nHitID = 0;
	Constraint * pBestConstraint = NULL;

	AnyConstraintSet::iterator cur(m_vConstraints.begin()), end(m_vConstraints.end());
	while ( cur != end ) {
		Constraint * pConstraint =  (*cur++).pConstraint;
		float fDistFromRay, fRayT;
		int nTempHitID = 0;
		if ( pConstraint->Intersect(ray, ptPosition, fDistFromRay, fRayT, nTempHitID) ) {
			if ( fRayT < fBestRayT ) {
				vBestHit = ptPosition;
				fBestHitDistFromRay = fDistFromRay;
				fBestRayT = fRayT;
				nHitID = nTempHitID;
				pBestConstraint = pConstraint;
			}
		}
	}
	
	if ( pBestConstraint != NULL && fBestHitDistFromRay < fHitThresh) 
		return pBestConstraint;
	return NULL;
}


Constraint * ConstraintSet::FindBestCross( const ViewState & eViewState, const Wml::Vector2f & v1, const Wml::Vector2f & v2, Wml::Vector3f & vHit3D )
{
	Constraint * pBestCross = NULL;
	float fNearest = std::numeric_limits<float>::max();

	AnyConstraintSet::iterator cur(m_vConstraints.begin()), end(m_vConstraints.end());
	while ( cur != end ) {
		Constraint * pConstraint =  (*cur++).pConstraint;

		Wml::Vector3f vTmpHit3D;
		bool bHit = pConstraint->TestCross( eViewState, v1, v2, vTmpHit3D );
		if ( bHit ) {
			float fDistSqr = (vTmpHit3D - CachedViewState().vEyePosition).SquaredLength();
			if ( fDistSqr < fNearest ) {
				fNearest = fDistSqr;
				vHit3D = vTmpHit3D;
				pBestCross = pConstraint;
			}
		}
	};

	return pBestCross;
}


void ConstraintSet::Clear( bool bClearPinned, bool bDeleteConstraints )
{
	VectorConstraintSet keepVec;
	PlaneTranslateConstraintSet keepPlaneTrans;
	AxisRotationConstraintSet keepAxisRotate;
	UniformScaleConstraintSet keepUniScale;
	AxisScaleConstraintSet keepAxisScale;

	VectorConstraintSet::iterator curv(m_vVectorConstraints.begin()), endv(m_vVectorConstraints.end());
	while ( curv != endv ) {
		VectorConstraint * pCons = *curv++;
		if ( ! bClearPinned && IsPinned(pCons) )
			keepVec.insert(pCons);
		else if (bDeleteConstraints) 
			delete pCons;
	}
	m_vVectorConstraints.clear();

	PlaneTranslateConstraintSet::iterator curt(m_vPlaneTranslateConstraints.begin()), endt(m_vPlaneTranslateConstraints.end());
	while ( curt != endt ) {
		PlaneTranslateConstraint * pCons = *curt++;
		if ( ! bClearPinned && IsPinned(pCons) )
			keepPlaneTrans.insert(pCons);
		else if (bDeleteConstraints) 
			delete pCons;
	}
	m_vPlaneTranslateConstraints.clear();

	AxisRotationConstraintSet::iterator 
		cura(m_vAxisRotationConstraints.begin()), 
		enda(m_vAxisRotationConstraints.end());
	while ( cura != enda ) {
		AxisRotationConstraint * pCons = *cura++;
		if ( ! bClearPinned && IsPinned(pCons) )
			keepAxisRotate.insert(pCons);
		else if (bDeleteConstraints) 
			delete pCons;
	}
	m_vAxisRotationConstraints.clear();

	UniformScaleConstraintSet::iterator 
		curu(m_vUniformScaleConstraints.begin()), 
		endu(m_vUniformScaleConstraints.end());
	while ( curu != endu ) {
		UniformScaleConstraint * pCons = *curu++;
		if ( ! bClearPinned && IsPinned(pCons) )
			keepUniScale.insert(pCons);
		else if (bDeleteConstraints) 
			delete pCons;
	}
	m_vUniformScaleConstraints.clear();

	AxisScaleConstraintSet::iterator 
		curs(m_vAxisScaleConstraints.begin()), 
		ends(m_vAxisScaleConstraints.end());
	while ( curs != ends ) {
		AxisScaleConstraint * pCons = *curs++;
		if ( ! bClearPinned && IsPinned(pCons) )
			keepAxisScale.insert(pCons);
		else if (bDeleteConstraints) 
			delete pCons;
	}
	m_vAxisScaleConstraints.clear();

	m_vConstraints.clear();


	// re-add pinned constraints
	curv = keepVec.begin(); endv = keepVec.end();
	while ( curv != endv ) 
		AddConstraint( *curv++ );
	curt = keepPlaneTrans.begin(); endt = keepPlaneTrans.end();
	while ( curt != endt ) 
		AddConstraint( *curt++ );
	cura = keepAxisRotate.begin(); enda = keepAxisRotate.end();
	while ( cura != enda ) 
		AddConstraint( *cura++ );
	curu = keepUniScale.begin(); endu = keepUniScale.end();
	while ( curu != endu ) 
		AddConstraint( *curu++ );
	curs = keepAxisScale.begin(); ends = keepAxisScale.end();
	while ( curs != ends ) 
		AddConstraint( *curs++ );
}



void ConstraintSet::TogglePinConstraint( Constraint * pCons )
{
	ConstraintInfo info;
	info.pConstraint = pCons;
	AnyConstraintSet::iterator found = m_vConstraints.find(info);
	if ( found != m_vConstraints.end() ) {
		_RMSInfo("Toggled Pin!\n");
		pCons->Pinned() = ! pCons->Pinned();

		// disappear un-pinned constraints
		if ( ! pCons->Pinned() )
			RemoveConstraint(pCons);

	} else {
		_RMSInfo("Toggle failed!\n");
	}

}

bool ConstraintSet::IsPinned( Constraint * pCons )
{
	ConstraintInfo info;
	info.pConstraint = pCons;
	AnyConstraintSet::iterator found = m_vConstraints.find(info);
	if ( found != m_vConstraints.end() ) {
		Constraint * pTmp = (*found).pConstraint;
		return pTmp->Pinned();
	}
	return false;
}


VectorConstraint * ConstraintSet::MakeVectorConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor )
{
	VectorConstraint * pCons = new VectorConstraint( 
		pTarget, axis.vOrigin, axis.vDirection );

	pCons->Color() = cColor;
	pCons->AxisType() = axis.eType;

	AddConstraint(pCons);
	return pCons;
}	
PlaneTranslateConstraint * ConstraintSet::MakePlaneTranslateConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor )
{
	PlaneTranslateConstraint * pCons = 
		(axis.m_bFrameSpecified) ?
			new PlaneTranslateConstraint( pTarget, axis.vFrame ) : 
			new PlaneTranslateConstraint( pTarget, axis.vOrigin, axis.vDirection );

	pCons->Color() = cColor;
	pCons->AxisType() = axis.eType;

	AddConstraint(pCons);
	return pCons;
}
AxisRotationConstraint * ConstraintSet::MakeAxisRotationConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, float fRadius, const Wml::ColorRGBA & cColor )
{
	AxisRotationConstraint * pCons = new AxisRotationConstraint( 
		pTarget, 0.4f, axis.vOrigin, axis.vDirection );

	pCons->SetRadius(fRadius);
	pCons->Color() = cColor;
	pCons->AxisType() = axis.eType;

	AddConstraint(pCons);
	return pCons;
}
UniformScaleConstraint * ConstraintSet::MakeUniformScaleConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor )
{
	UniformScaleConstraint * pCons = new UniformScaleConstraint( pTarget, axis.vOrigin );

	pCons->Color() = cColor;
	pCons->AxisType() = axis.eType;

	AddConstraint(pCons);
	return pCons;
}
AxisScaleConstraint * ConstraintSet::MakeAxisScaleConstraint( ConstraintTarget * pTarget, const ConstraintAxis & axis, const Wml::ColorRGBA & cColor )
{
	AxisScaleConstraint * pCons = new AxisScaleConstraint( 
		pTarget, axis.vOrigin, axis.vDirection );

	pCons->Color() = cColor;
	pCons->AxisType() = axis.eType;

	AddConstraint(pCons);
	return pCons;
}









void ConstraintSet::AddConstraint( VectorConstraint * pVec )
{
	ConstraintInfo info;
	info.pConstraint = pVec;
	m_vConstraints.insert(info);

	m_vVectorConstraints.insert(pVec);
}

void ConstraintSet::AddConstraint( PlaneTranslateConstraint * pVec )
{
	ConstraintInfo info;
	info.pConstraint = pVec;
	m_vConstraints.insert(info);

	m_vPlaneTranslateConstraints.insert(pVec);
}

void ConstraintSet::AddConstraint( AxisRotationConstraint * pAxisRot )
{
	ConstraintInfo info;
	info.pConstraint = pAxisRot;
	m_vConstraints.insert(info);

	m_vAxisRotationConstraints.insert(pAxisRot);
}
void ConstraintSet::AddConstraint( UniformScaleConstraint * pUniScale )
{
	ConstraintInfo info;
	info.pConstraint = pUniScale;
	m_vConstraints.insert(info);

	m_vUniformScaleConstraints.insert(pUniScale);
}
void ConstraintSet::AddConstraint( AxisScaleConstraint * pAxisScale )
{
	ConstraintInfo info;
	info.pConstraint = pAxisScale;
	m_vConstraints.insert(info);

	m_vAxisScaleConstraints.insert(pAxisScale);
}


void ConstraintSet::RemoveConstraint( Constraint * pConstraint )
{
	ConstraintInfo info;
	info.pConstraint = pConstraint;

	m_vConstraints.erase(info);
	m_vVectorConstraints.erase((VectorConstraint *)pConstraint);	
	m_vPlaneTranslateConstraints.erase((PlaneTranslateConstraint *)pConstraint);
	m_vAxisRotationConstraints.erase((AxisRotationConstraint *)pConstraint);
	m_vUniformScaleConstraints.erase((UniformScaleConstraint *)pConstraint);
	m_vAxisScaleConstraints.erase((AxisScaleConstraint *)pConstraint);
}




void ConstraintSet::FindConstraintsByTarget( ConstraintTarget * pTarget, std::vector<Constraint *> & vConstraints )
{
	vConstraints.resize(0);
	AnyConstraintSet::iterator cur(m_vConstraints.begin()), end(m_vConstraints.end());
	while ( cur != end ) {
		Constraint * pTmp = (*cur++).pConstraint;
		if ( *(pTmp->Target()) == (*pTarget) )
			vConstraints.push_back(pTmp);
	}
}

VectorConstraint * ConstraintSet::FindVectorConstraint( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bPinnedOnly )
{
	static const float fThresh = 0.0001f;

	VectorConstraintSet::iterator cur(m_vVectorConstraints.begin()), end(m_vVectorConstraints.end());
	while ( cur != end ) {
		VectorConstraint * pCons = *cur++;
		if ( bPinnedOnly && ! pCons->Pinned() )
			continue;
		float fDot = pCons->Axis().Dot( vDirection );
		float fDelta = (pCons->Origin() - vOrigin).SquaredLength();
		if ( fabs( 1.0f - fabs(fDot) ) < fThresh && fDelta < fThresh)
			return pCons;
	}
	return NULL;
}
AxisRotationConstraint * ConstraintSet::FindAxisRotationConstraint( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bPinnedOnly )
{
	static const float fThresh = 0.0001f;

	AxisRotationConstraintSet::iterator cur(m_vAxisRotationConstraints.begin()), end(m_vAxisRotationConstraints.end());
	while ( cur != end ) {
		AxisRotationConstraint * pCons = *cur++;
		if ( bPinnedOnly && ! pCons->Pinned() )
			continue;
		float fDot = pCons->Axis().Dot( vDirection );
		float fDelta = (pCons->Origin() - vOrigin).SquaredLength();
		if ( fabs( 1.0f - fabs(fDot) ) < fThresh && fDelta < fThresh)
			return pCons;
	}
	return NULL;
}
PlaneTranslateConstraint * ConstraintSet::FindPlaneTranslateConstraint( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bPinnedOnly )
{
	static const float fThresh = 0.0001f;

	PlaneTranslateConstraintSet::iterator cur(m_vPlaneTranslateConstraints.begin()), end(m_vPlaneTranslateConstraints.end());
	while ( cur != end ) {
		PlaneTranslateConstraint * pCons = *cur++;
		if ( bPinnedOnly && ! pCons->Pinned() )
			continue;
		float fDot = pCons->Axis().Dot( vDirection );
		float fDelta = (pCons->Origin() - vOrigin).SquaredLength();
		if ( fabs( 1.0f - fabs(fDot) ) < fThresh && fDelta < fThresh)
			return pCons;
	}
	return NULL;
}
UniformScaleConstraint * ConstraintSet::FindUniformScaleConstraint( ConstraintTarget * pTarget )
{
	UniformScaleConstraintSet::iterator cur(m_vUniformScaleConstraints.begin()), end(m_vUniformScaleConstraints.end());
	while ( cur != end ) {
		UniformScaleConstraint * pCons = *cur++;
		if ( *(pCons->Target()) == *pTarget )
			return pCons;
	}
	return NULL;
}


bool ConstraintSet::FindConstraints( const Wml::Vector3f & vOrigin, const Wml::Vector3f & vDirection, bool bSkipPinned, std::vector<Constraint *> & vFound )
{
	static const float fThresh = 0.0001f;
	vFound.resize(0);

	VectorConstraintSet::iterator curv(m_vVectorConstraints.begin()), endv(m_vVectorConstraints.end());
	while ( curv != endv ) {
		VectorConstraint * pCons = *curv++;
		if ( bSkipPinned && IsPinned(pCons) )
			continue;
		float fDot = pCons->Axis().Dot( vDirection );
		float fDelta = (pCons->Origin() - vOrigin).SquaredLength();
		if ( fabs( 1.0f - fabs(fDot) ) < fThresh && fDelta < fThresh)
			vFound.push_back(pCons);
	}

	PlaneTranslateConstraintSet::iterator curt(m_vPlaneTranslateConstraints.begin()), endt(m_vPlaneTranslateConstraints.end());
	while ( curt != endt ) {
		PlaneTranslateConstraint * pCons = *curt++;
		if ( bSkipPinned && IsPinned(pCons) )
			continue;
		float fDot = pCons->Axis().Dot( vDirection );
		float fDelta = (pCons->Origin() - vOrigin).SquaredLength();
		if ( fabs( 1.0f - fabs(fDot) ) < fThresh && fDelta < fThresh)
			vFound.push_back(pCons);
	}

	AxisRotationConstraintSet::iterator cura(m_vAxisRotationConstraints.begin()), enda(m_vAxisRotationConstraints.end());
	while ( cura != enda ) {
		AxisRotationConstraint * pCons = *cura++;
		if ( bSkipPinned && IsPinned(pCons) )
			continue;
		float fDot = pCons->Axis().Dot( vDirection );
		float fDelta = (pCons->Origin() - vOrigin).SquaredLength();
		if ( fabs( 1.0f - fabs(fDot) ) < fThresh && fDelta < fThresh)
			vFound.push_back(pCons);
	}

	return vFound.empty();
}






void ConstraintSet::AppendHistoryCheckpoint()
{
	HistoryEvent event;

	VectorConstraintSet::iterator curv(m_vVectorConstraints.begin()), endv(m_vVectorConstraints.end());
	while ( curv != endv ) {
		VectorConstraint * pCons = *curv++;
		
		ConstraintAxis axis(pCons->AxisType(), pCons->Origin(), pCons->Axis());
		axis.m_bPinned = pCons->Pinned();
		axis.m_eRestoreType = pCons->GetType();
		axis.m_vColor = pCons->Color();
		event.vConstraints.push_back(axis);
	}

	PlaneTranslateConstraintSet::iterator curt(m_vPlaneTranslateConstraints.begin()), endt(m_vPlaneTranslateConstraints.end());
	while ( curt != endt ) {
		PlaneTranslateConstraint * pCons = *curt++;

		ConstraintAxis axis(pCons->AxisType(), pCons->Origin(), pCons->Axis());
		axis.m_bPinned = pCons->Pinned();
		axis.m_eRestoreType = pCons->GetType();
		axis.m_vColor = pCons->Color();
		event.vConstraints.push_back(axis);
	}

	AxisRotationConstraintSet::iterator cura(m_vAxisRotationConstraints.begin()), enda(m_vAxisRotationConstraints.end());
	while ( cura != enda ) {
		AxisRotationConstraint * pCons = *cura++;

		ConstraintAxis axis(pCons->AxisType(), pCons->Origin(), pCons->Axis());
		axis.m_bPinned = pCons->Pinned();
		axis.m_eRestoreType = pCons->GetType();
		axis.m_vColor = pCons->Color();
		axis.m_fRadius = pCons->GetRadius();
		event.vConstraints.push_back(axis);
	}

	UniformScaleConstraintSet::iterator curu(m_vUniformScaleConstraints.begin()), endu(m_vUniformScaleConstraints.end());
	while ( curu != endu ) {
		UniformScaleConstraint * pCons = *curu++;

		ConstraintAxis axis(pCons->AxisType(), pCons->Origin(), pCons->Axis());
		axis.m_bPinned = pCons->Pinned();
		axis.m_eRestoreType = pCons->GetType();
		axis.m_vColor = pCons->Color();
		event.vConstraints.push_back(axis);
	}

	AxisScaleConstraintSet::iterator curs(m_vAxisScaleConstraints.begin()), ends(m_vAxisScaleConstraints.end());
	while ( curs != ends ) {
		AxisScaleConstraint * pCons = *curs++;

		ConstraintAxis axis(pCons->AxisType(), pCons->Origin(), pCons->Axis());
		axis.m_bPinned = pCons->Pinned();
		axis.m_eRestoreType = pCons->GetType();
		axis.m_vColor = pCons->Color();
		event.vConstraints.push_back(axis);
	}

	if ( event.vConstraints.size() > 0 )
		m_vHistory.push_back(event);
}


bool ConstraintSet::SetHistoryState( unsigned int nState, ConstraintTarget * pTarget )
{
	size_t nEvents = m_vHistory.size();
	if ( nState >= nEvents )
		return false;
	HistoryEvent & event = m_vHistory[nEvents - nState - 1];
	size_t nCount = event.vConstraints.size();

	Wml::Vector3f vOrigin = pTarget->GetFrame().Origin();

	for ( unsigned int i = 0; i < nCount; ++i ) {
		ConstraintAxis axis = event.vConstraints[i];	// copy!

		// reset origin
		axis.vOrigin = vOrigin;
		axis.vFrame.Origin() = vOrigin;

		switch ( axis.m_eRestoreType ) {

			case Constraint::AxisTranslate: {
				VectorConstraint * pCons = MakeVectorConstraint( pTarget, axis, axis.m_vColor );
				}
				break;

			case Constraint::PlaneTranslate: {
				PlaneTranslateConstraint * pCons = MakePlaneTranslateConstraint( pTarget, axis, axis.m_vColor );
				}
				break;

			case Constraint::AxisRotate: {
				AxisRotationConstraint * pCons = MakeAxisRotationConstraint( pTarget, axis, axis.m_fRadius, axis.m_vColor );
				}
				break;

			case Constraint::UniformScale: {
				UniformScaleConstraint * pCons = MakeUniformScaleConstraint( pTarget, axis, axis.m_vColor );
				}
				break;

			case Constraint::AxisScale: {
				AxisScaleConstraint * pCons = MakeAxisScaleConstraint( pTarget, axis, axis.m_vColor );
				}
				break;
		}
	}

	return true;
}





void PivotSet::AddPivot( PivotSO * pPivot )
{
	m_vPivots.push_back(pPivot);
}

void PivotSet::DeletePivot( PivotSO * pPivot )
{
	std::vector<PivotSO *>::iterator cur(m_vPivots.begin()), end(m_vPivots.end());
	while ( cur != end ) {
		if ( *cur == pPivot ) {
			m_vPivots.erase(cur);
			return;
		}
		++cur;
	}
}

void PivotSet::ClearPivots()
{
	m_vPivots.resize(0);
}



void PivotSet::Render(const ViewState & viewState)
{
	size_t nCount = m_vPivots.size();
	for ( unsigned int i = 0; i < nCount; ++i )
		m_vPivots[i]->Render(viewState);
}

PivotSO * PivotSet::Intersect( const Wml::Ray3f & ray, Wml::Vector3f & ptPosition, float & fDepth )
{
	float fNearest = std::numeric_limits<float>::max();
	PivotSO * pHit = NULL;

	size_t nCount = m_vPivots.size();
	for ( unsigned int i = 0; i < nCount; ++i ) {
		Wml::Vector3f vHit;
		if ( m_vPivots[i]->Intersect( ray, vHit ) ) {
			float fDistSqr = (vHit - ray.Origin()).SquaredLength();
			if ( fDistSqr < fNearest ) {
				fNearest = fDistSqr;
				ptPosition = vHit;
				pHit = m_vPivots[i];
			}
		}
	}
	if ( pHit != NULL )
		fDepth = (float)sqrt(fNearest);
	return pHit;
}


void PivotSet::ClearTargets()
{
	size_t nCount = m_vPivots.size();
	for ( unsigned int i = 0; i < nCount; ++i ) 
		m_vPivots[i]->SetTarget(NULL);
}

