// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "GestureEngine.h"

#include <WmlDistVec2Lin2.h>
#include <WmlIntrLin2Box2.h>
#include <WmlExtDistVec2Box2.h>
#include <WmlContBox2.h>

#include <gl/GL.h>
#include <gl/GLU.h>

using namespace rmsui;



GestureEngine::GestureEngine()
{
	m_pCurrentStroke = NULL;
}

GestureEngine::~GestureEngine()
{
}


void GestureEngine::RenderCurrentStroke(unsigned int nScreenWidth, unsigned int nScreenHeight)
{
	if ( m_pCurrentStroke == NULL )
		return;

	glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D( 0.0f, (float)nScreenWidth, 0.0f, (float)nScreenHeight );
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor3f(0.0f, 0.0f, 1.0f);
	glLineWidth(4.0f);

	m_pCurrentStroke->DrawGL();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glPopAttrib();
}

IInputEventListener::EventState GestureEngine::ProcessInputEvent( const InputEventData & event )
{
	Wml::Vector2f vPoint( (float)event.nMouseX, (float)event.nMouseY );

	if ( event.LeftButton.bDown == true ) {

		// if left button was not down...
		if ( event.LeftButton.bWasDown == false ) {

			// create a stroke
			StartStroke(vPoint);

			return IInputEventListener::StartCapture;

		// otherwise append to the current stroke (if one is being created)
		} else if ( m_pCurrentStroke && AppendToStroke(vPoint) != false ) {
			return IInputEventListener::Consumed;
		}

	} else if ( event.LeftButton.bWasDown == true ) {

		// if we are in a stroke, so end it. Returns false if no HUD items care about stroke
		if ( m_pCurrentStroke != NULL ) {
			EndStroke(vPoint);

			// [RMS: fixes crash bug w/doubleclick, which sends two mouse-up events in a row but
			//   m_pCurrentStroke is not destroyed the first time so it is != NULL and we send
			//   a second EndCapture which screws up InputEventStream
			if ( m_pCurrentStroke != NULL ) {
				delete m_pCurrentStroke;
				m_pCurrentStroke = NULL;
			}

			return IInputEventListener::EndCapture;
		}

	} 

	return IInputEventListener::Ignored;
}



void GestureEngine::StartStroke( const Wml::Vector2f & vStart )
{
	if ( m_pCurrentStroke != NULL ) {
		delete m_pCurrentStroke;
		m_pCurrentStroke = NULL;
	}
	m_pCurrentStroke = new rmsui::GestureStroke();
	m_pCurrentStroke->AppendVertex(vStart);
}

bool GestureEngine::AppendToStroke( const Wml::Vector2f & vAppend )
{
	if (!m_pCurrentStroke ) {
		ASSERT(m_pCurrentStroke);
		return false;
	}
	m_pCurrentStroke->AppendVertex(vAppend);

	return true;  // always true?
}

bool GestureEngine::EndStroke( const Wml::Vector2f & vEnd )
{
	ASSERT(m_pCurrentStroke);
	m_pCurrentStroke->AppendVertex(vEnd);

	return CheckCurrentStroke();
}

// RMS TODO: this is where we detect gestures and send out as events...
bool GestureEngine::CheckCurrentStroke()
{
	ASSERT(m_pCurrentStroke);

	bool bConsumed = false;
	bool bTriedGeneralStroke = false;

	if ( m_pCurrentStroke->IsPointStroke() ) {
		rmsui::GesturePointStroke pointStroke;
		m_pCurrentStroke->GetPointStroke(&pointStroke);

		bConsumed = NotifyPointStroke( & pointStroke );

	} else if ( m_pCurrentStroke->IsLineStroke() ) {
		rmsui::GestureLineStroke lineStroke;
		m_pCurrentStroke->GetLineStroke(&lineStroke);

		bConsumed = NotifyLineStroke( & lineStroke );
	} else if ( m_pCurrentStroke->IsVStroke() ) {
		rmsui::GestureVStroke vStroke;
		m_pCurrentStroke->GetVStroke(&vStroke);

		bConsumed = NotifyVStroke( & vStroke );
	} else if ( m_pCurrentStroke->IsScribbleStroke() ) {
		rmsui::GestureScribbleStroke scribbleStroke;
		m_pCurrentStroke->GetScribbleStroke(&scribbleStroke);

		bConsumed = NotifyScribbleStroke( & scribbleStroke );
	} else {
		bConsumed = NotifyGeneralStroke( m_pCurrentStroke );
		bTriedGeneralStroke = true;
	}

	if ( ! bConsumed && ! bTriedGeneralStroke)
		bConsumed = NotifyGeneralStroke( m_pCurrentStroke );

	return bConsumed;
}


bool GestureEngine::NotifyLineStroke( GestureLineStroke * pStroke )
{
	std::vector<IGestureEventListener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
	bool bConsumed = false;
	while ( cur != end && ! bConsumed ) {
		bConsumed = (*cur++)->OnLineStroke(pStroke);
	}
	return bConsumed;
}

bool GestureEngine::NotifyPointStroke( GesturePointStroke * pStroke )
{
	std::vector<IGestureEventListener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
	bool bConsumed = false;
	while ( cur != end && ! bConsumed ) {
		bConsumed = (*cur++)->OnPointStroke(pStroke);
	}
	return bConsumed;
}

bool GestureEngine::NotifyVStroke( GestureVStroke * pStroke )
{
	std::vector<IGestureEventListener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
	bool bConsumed = false;
	while ( cur != end && ! bConsumed ) {
		bConsumed = (*cur++)->OnVStroke(pStroke);
	}
	return bConsumed;
}

bool GestureEngine::NotifyScribbleStroke( GestureScribbleStroke * pStroke )
{
	std::vector<IGestureEventListener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
	bool bConsumed = false;
	while ( cur != end && ! bConsumed ) {
		bConsumed = (*cur++)->OnScribbleStroke(pStroke);
	}
	return bConsumed;
}


bool GestureEngine::NotifyGeneralStroke( GestureStroke * pStroke )
{
	std::vector<IGestureEventListener *>::iterator cur(m_vListeners.begin()), end(m_vListeners.end());
	bool bConsumed = false;
	while ( cur != end && ! bConsumed ) {
		bConsumed = (*cur++)->OnGeneralStroke(pStroke);
	}
	return bConsumed;
}