// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "GestureStrokes.h"

#include <WmlDistVec2Lin2.h>
#include <WmlIntrLin2Box2.h>
#include <WmlExtDistVec2Box2.h>
#include <WmlContBox2.h>

#include <gl/GL.h>
#include <gl/GLU.h>

using namespace rmsui;

/*
 * GestureStroke
 */

GestureStroke::GestureStroke()
{
}

GestureStroke::~GestureStroke()
{
}

GestureStroke::GestureStroke( const std::vector<Wml::Vector2f> & vVertices )
{	
	m_vVertices = vVertices;
}


void GestureStroke::DrawGL()
{
	glBegin(GL_LINE_STRIP);
	size_t nCount = m_vVertices.size();
	for ( unsigned int i = 0; i < nCount; ++i )
		glVertex2fv( m_vVertices[i] );
	glEnd();
}

Wml::Box2f GestureStroke::GetBoundingBox()
{
	return Wml::ContOrientedBox( (int)m_vVertices.size(), &m_vVertices[0] );
}



void GestureStroke::AppendVertex( const Wml::Vector2f & vVertex, bool bSkipDuplicates ) 
{
	if (bSkipDuplicates && ! m_vVertices.empty() ) {
		float fDistSqr = ( m_vVertices.back() - vVertex ).SquaredLength();
		if ( fDistSqr > 0.01f )				// this is in pixel space...
			m_vVertices.push_back(vVertex);
	} else
		m_vVertices.push_back(vVertex);
}




bool GestureStroke::IsMultipleLoopsStroke()
{
	Wml::Box2f bbox( GetBoundingBox() );

	Wml::Vector2f vCenter( bbox.Center()) ;

	float fAccumAngle = 0.0f;

	size_t nSize = m_vVertices.size();
	for (unsigned int i = 1; i < nSize; ++i) {

		Wml::Vector2f vLeft( m_vVertices[i-1] - vCenter );
		vLeft.Normalize();
		Wml::Vector2f vRight( m_vVertices[i] - vCenter );
		vRight.Normalize();

		float fAngle = std::acos( vLeft.Dot(vRight) );
		int nSign = (fAngle > 0) ? 1 : 0;
		fAccumAngle += fAngle;
	}

	if ( fabs(fAccumAngle) > Wml::Mathf::PI*4 )
		return true;
	return false;
}



// check if basically closed (?)
bool GestureStroke::IsLoopStroke()
{
	return false;
}


bool GestureStroke::IsLineStroke()
{
	size_t nSize = m_vVertices.size();
	if (nSize < 2)
		return false;

	return IsLine( 0, (unsigned int)nSize-1 );
}


bool GestureStroke::GetLineStroke( GestureLineStroke * pStroke )
{
	if (!pStroke) {
		ASSERT(pStroke);
		return false;
	}
	if ( m_vVertices.size() < 2 )
		return false;

	pStroke->EndPoint1() = m_vVertices.front();
	pStroke->EndPoint2() = m_vVertices.back();

	return true;
}




bool GestureStroke::IsPointStroke()
{
	if ( m_vVertices.empty() )
		return false;

	// [RMS: should have some idea of pixel space to do this properly...]
	float fMaxDistSqr = 10.0f * 10.0f;

	size_t nCount = m_vVertices.size();
	for ( unsigned int i = 1; i < nCount; ++i ) {
		float fDistSqr = (m_vVertices[i] - m_vVertices[0] ).SquaredLength();
		if ( fDistSqr > fMaxDistSqr )
			return false;
	}
	return true;
}

bool GestureStroke::GetPointStroke( GesturePointStroke * pStroke )
{
	if (!pStroke) {
		ASSERT(pStroke);
		return false;
	}
	if ( m_vVertices.empty() )
		return false;

	pStroke->Point() = m_vVertices.front();

	return true;
}



bool GestureStroke::IsVStroke()
{
	size_t nSize = m_vVertices.size();
	if ( nSize < 3 )
		return false;

	// first and last segments should point in same direction
	Wml::Vector2f vSegStart( m_vVertices[0] - m_vVertices[1] );
	Wml::Vector2f vSegEnd( m_vVertices[nSize-1] - m_vVertices[nSize-2] );
	if ( vSegStart.Dot(vSegEnd) < 0 )
		return false;

	// should also both point 'up'
	vSegStart.Normalize();
	if ( vSegStart.Dot( Wml::Vector2f::UNIT_Y ) < 0.5f )
		return false;
	vSegEnd.Normalize();
	if ( vSegEnd.Dot( Wml::Vector2f::UNIT_Y ) < 0.5f )
		return false;

	// ok, it might be a V. find point with smallest angle...
	unsigned int nLargest = 0;
	float fLargestAngle = 999999.0f;
	for ( unsigned int i = 1; i < nSize-1; ++i ) {
		Wml::Vector2f vSeg1( m_vVertices[i-1] - m_vVertices[i] );
		Wml::Vector2f vSeg2( m_vVertices[i+1] - m_vVertices[i] );
		float fLength1 = vSeg1.Length();
		float fLength2 = vSeg2.Length();
		if ( fLength1 < 0.01f || fLength2 < 0.01f )
			continue;
		float fDot = vSeg1.Dot(vSeg2) / ( fLength1 * fLength2 );
		float fAngle = (float)acos(fDot);
		if ( fAngle < fLargestAngle ) {
			fLargestAngle = fAngle;
			nLargest = i;
		}
	}

	if (nLargest == 0)
		return false;

	// both sides should be about the same length
	Wml::Vector2f vSide1( m_vVertices[0] - m_vVertices[nLargest] );
	Wml::Vector2f vSide2( m_vVertices[nSize-1] - m_vVertices[nLargest] );
	float fLength1 = vSide1.Length();
	float fLength2 = vSide2.Length();
	float fRatio = fLength1 / fLength2;
	if ( fRatio < 0.67f || fRatio > 1.5f )
		return false;

	// ok, now the segments before and after the largest angle should be basically straight...
	if ( IsLine( 0, nLargest ) && IsLine( nLargest, (unsigned int)nSize-1 ) )
		return true;
	else
		return false;
}

bool GestureStroke::GetVStroke( GestureVStroke * pStroke )
{
	if (!pStroke) {
		ASSERT(pStroke);
		return false;
	}
	if ( m_vVertices.size() < 3 )
		return false;	

	pStroke->Start() = m_vVertices.front();
	pStroke->End() = m_vVertices.back();

	return true;
}




// basically linear segments with occasional sharp angles
bool GestureStroke::IsScribbleStroke()
{
	size_t nSize = m_vVertices.size();
	unsigned int nSharpTransitions = 0;
	unsigned int nLinearSegments = 0;
	unsigned int nNonLinearSegments = 0;

	unsigned int nStartCurSeg = 0;
	unsigned int nEndCurSeg = 0;

	float fCosThreshold = cos(30.0f * Wml::Mathf::DEG_TO_RAD );

	for (unsigned int i = 1; i < nSize-1; ++i) {

		Wml::Vector2f vLeft( m_vVertices[i-1] - m_vVertices[i] );
		vLeft.Normalize(); 
		Wml::Vector2f vRight( m_vVertices[i+1] - m_vVertices[i] );
		vRight.Normalize();

		// compute cos angle between them
		float fCosAngle = vLeft.Dot(vRight);

		if (fCosAngle > fCosThreshold || i == nSize-2 ) {
			nSharpTransitions++;

			// find line between this point and start of cur seg
			Wml::Segment2f segLine;
			segLine.Origin() = m_vVertices[nStartCurSeg];
			segLine.Direction() = m_vVertices[i] - segLine.Origin();
			float fLineLenSqr = segLine.Direction().SquaredLength();

			// project points onto line. if any dist > linelen / 4, this is not a linear segment
			fLineLenSqr /= 16.0f;
			bool bLinear = true;
			for (unsigned int j = nStartCurSeg+1; j < i && bLinear; ++j) {
				float fDistSqr = SqrDistance( m_vVertices[j], segLine );
				if (fDistSqr > fLineLenSqr)
					bLinear = false;
			}

			if (bLinear)
				nLinearSegments++;
			else
				nNonLinearSegments++;

			nStartCurSeg = i;
		}
	}


	if ( nSharpTransitions > 4 && nLinearSegments > nNonLinearSegments)
		return true;
	else
		return false;
}



bool GestureStroke::GetScribbleStroke( GestureScribbleStroke * pStroke )
{
	if (!pStroke) {
		ASSERT(pStroke);
		return false;
	}
	if ( m_vVertices.size() < 6 )
		return false;	

	pStroke->Bounds() = GetBoundingBox();
	return true;
}



bool GestureStroke::IsLine( unsigned int nStartVertex, unsigned int nEndVertex )
{
	Wml::Segment2f segLine;
	segLine.Origin() = m_vVertices[nStartVertex];
	segLine.Direction() = m_vVertices[nEndVertex] - segLine.Origin();
	float fLineLenSqr = segLine.Direction().SquaredLength();
	fLineLenSqr /= 200.0f;

	int nLinearFailed = 0;
	for (unsigned int i = nStartVertex + 1; i < nEndVertex - 1; ++i) {
		float fDistSqr = SqrDistance( m_vVertices[i], segLine );
		if (fDistSqr > fLineLenSqr)
			++nLinearFailed;
	}

	if ( (float)nLinearFailed / (float)(nEndVertex-nStartVertex) < 0.1f )
		return true;
	else
		return false;
}




/*
 * GestureLineStroke
 */

GestureLineStroke::GestureLineStroke( )
	: m_pt1(Wml::Vector2f::ZERO), m_pt2(Wml::Vector2f::ZERO)
{
}


GestureLineStroke::GestureLineStroke( const Wml::Vector2f & pt1, const Wml::Vector2f & pt2 )
	: m_pt1(pt1), m_pt2(pt2)
{
}

GestureLineStroke::~GestureLineStroke()
{
}



/*
 * GesturePointStroke
 */



GesturePointStroke::GesturePointStroke( )
	: m_pt(Wml::Vector2f::ZERO)
{
}


GesturePointStroke::GesturePointStroke( const Wml::Vector2f & pt )
	: m_pt(pt)
{
}

GesturePointStroke::~GesturePointStroke()
{
}



/*
 * GestureVStroke
 */
GestureVStroke::GestureVStroke( )
	: m_ptStart(Wml::Vector2f::ZERO), m_ptEnd(Wml::Vector2f::ZERO)
{
}


GestureVStroke::GestureVStroke( const Wml::Vector2f & ptStart, const Wml::Vector2f & ptEnd )
	: m_ptStart(ptStart), m_ptEnd(ptEnd)
{
}

GestureVStroke::~GestureVStroke()
{
}


GestureScribbleStroke::GestureScribbleStroke()
{
}

GestureScribbleStroke::GestureScribbleStroke( const Wml::Box2f & BoundingBox )
	: m_boundingBox(BoundingBox)
{
}

GestureScribbleStroke::~GestureScribbleStroke()
{
}


