// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2005-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#include "StdAfx.h"
#include "RenderBuffer.h"

#include "rmsdebug.h"

using namespace rmsui;

RenderBuffer::RenderBuffer()
{
	m_pCDC = NULL;
	m_pBitmap = NULL;
	m_hGLRC = NULL;

	m_nWidth = m_nHeight = 0;
}


RenderBuffer::~RenderBuffer()
{
	if ( m_hGLRC ) {
		HGLRC hCurGLRC = wglGetCurrentContext();
		HDC hCurDC = wglGetCurrentDC();

		if ( hCurGLRC != m_hGLRC )
			wglMakeCurrent( m_pCDC->GetSafeHdc(), m_hGLRC);
		wglDeleteContext( m_hGLRC );
		m_hGLRC = NULL;

		wglMakeCurrent( hCurDC, hCurGLRC );
	}

	if ( m_pCDC ) {
		m_pCDC->SelectObject(m_pOldDCBitmap);
		delete m_pCDC;
	}

	if ( m_pBitmap )
		delete m_pBitmap;

}

bool RenderBuffer::Initialize(unsigned int nWidth, unsigned int nHeight)
{
	m_pCDC = new CDC();
	m_pCDC->CreateCompatibleDC( NULL );

	m_pBitmap = new CBitmap();
	m_pBitmap->CreateBitmap( nWidth, nHeight, 1, 32, NULL );

	m_pOldDCBitmap = m_pCDC->SelectObject(m_pBitmap);

	if ( SetPixelFormat( m_pCDC ) == FALSE )
		return false;

	// create OpenGL RC
	m_hGLRC = wglCreateContext(m_pCDC->GetSafeHdc());

	ASSERT(m_hGLRC);

	m_nWidth = nWidth;
	m_nHeight = nHeight;
	return true;
}

void RenderBuffer::EnableGLContext(bool bSaveContext)
{
	if ( bSaveContext ) {
		m_hSavedContextDC = wglGetCurrentDC();
		m_hSavedContextGLRC = wglGetCurrentContext();
	}

	if ( m_pCDC && m_hGLRC )
		wglMakeCurrent( m_pCDC->GetSafeHdc(), m_hGLRC );
}

void RenderBuffer::DisableGLContext(bool bRestoreContext)
{
	wglMakeCurrent(NULL, NULL);

	if ( bRestoreContext ) 
		wglMakeCurrent( m_hSavedContextDC, m_hSavedContextGLRC );
}


BOOL RenderBuffer::SetPixelFormat( CDC * pDC )
{
	CBitmap* pBitmap = pDC->GetCurrentBitmap() ;

	BITMAP bmInfo ;
	pBitmap->GetObject(sizeof(BITMAP), &bmInfo) ;

	// sanity checks...
	ASSERT(bmInfo.bmPlanes == 1) ;
	ASSERT((bmInfo.bmBitsPixel == 8) || 
		(bmInfo.bmBitsPixel == 16) ||
		(bmInfo.bmBitsPixel == 24) ||
		(bmInfo.bmBitsPixel == 32) );

	// define a default desired video mode (pixel format)
	static PIXELFORMATDESCRIPTOR pfd = 
	{
		sizeof(PIXELFORMATDESCRIPTOR),	// size of this pfd
			1,								// version number
			PFD_DRAW_TO_BITMAP |			// support window
			PFD_SUPPORT_OPENGL,				// support OpenGL
			PFD_TYPE_RGBA,                  // RGBA type
			32,                             // 32-bit color depth
			0, 0, 0, 0, 0, 0,               // color bits ignored
			8,                              // 8-bit alpha buffer
			0,                              // shift bit ignored
			0,                              // no accumulation buffer
			0, 0, 0, 0,                     // accum bits ignored
			16,                             // 16-bit z-buffer
			0,                              // no stencil buffer
			0,                              // no auxiliary buffer
			PFD_MAIN_PLANE,                 // main layer
			0,                              // reserved
			0, 0, 0                         // layer masks ignored
	};


	// try to find a compatible PixelFormat
	ASSERT(pDC != NULL);
	int pixelformat;
	if ( (pixelformat = ChoosePixelFormat(pDC->GetSafeHdc(), &pfd)) == 0 ){
		AfxMessageBox("ChoosePixelFormat failed");
		return FALSE;
	}

	// try to switch to selected PixelFormat    
	if (::SetPixelFormat(pDC->GetSafeHdc(), pixelformat, &pfd) == FALSE){

		// failed, try a default?
		pixelformat = 1;	
		if (DescribePixelFormat(pDC->GetSafeHdc(), pixelformat, sizeof(PIXELFORMATDESCRIPTOR), &pfd)==0){
			return FALSE;
		}
	}
	return TRUE;
}



void RenderBuffer::GetBuffer( std::vector<unsigned char> & vBuffer, 
							  const unsigned int nChannelMap[3],
							  bool bWantAlpha )
{
	ASSERT(!bWantAlpha);
	vBuffer.resize(0);

	unsigned char nRGB[3];
	// loop backwards for stupid windows "top is bottom"
	for ( int yi = m_nHeight-1; yi >= 0; --yi ) {
		for ( unsigned int xi = 0; xi  < m_nWidth; ++xi ) {
			COLORREF c = m_pCDC->GetPixel(xi, yi);
			nRGB[0] = GetRValue(c);
			nRGB[1] = GetGValue(c);
			nRGB[2] = GetBValue(c);
			unsigned char nAlpha = ((unsigned char)((c)>>24));

			vBuffer.push_back( nRGB[nChannelMap[0]] );
			vBuffer.push_back( nRGB[nChannelMap[1]] );
			vBuffer.push_back( nRGB[nChannelMap[2]] );
			if ( bWantAlpha ) {
				if (nAlpha != 0)
					_RMSInfo("%d\n", nAlpha);
				vBuffer.push_back(255);
			}
		}
	}
}


void RenderBuffer::GetGLBuffer( std::vector<unsigned char> & vBuffer, 
							    bool bWantAlpha )
{
	vBuffer.resize( (bWantAlpha) ? m_nWidth*m_nHeight*4 : m_nWidth*m_nHeight*3 );
	if ( bWantAlpha )
		glReadPixels(0, 0, m_nWidth, m_nHeight, GL_RGBA, GL_UNSIGNED_BYTE, &vBuffer[0] );
	else
		glReadPixels(0, 0, m_nWidth, m_nHeight, GL_RGB, GL_UNSIGNED_BYTE, &vBuffer[0] );
}

void RenderBuffer::GetDepthBuffer( std::vector<float> & vBuffer, bool bScale )
{
	vBuffer.resize( m_nWidth * m_nHeight );

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadPixels( 0, 0, m_nWidth, m_nHeight, GL_DEPTH_COMPONENT, GL_FLOAT, &vBuffer[0] );

	if ( bScale ) {
		float fMax = -10000.0f, fMin = 10000.0f;
		for ( unsigned int i = 0; i < m_nWidth*m_nHeight; ++i ) {
			if ( vBuffer[i] < fMin )
				fMin = vBuffer[i];
			if ( vBuffer[i] > fMax )
				fMax = vBuffer[i];
		}
		float fMinZ = fMin;
		float fRange = 1.0f / (fMax - fMin);

		for ( unsigned int i = 0; i < m_nWidth*m_nHeight; ++i ) {
			vBuffer[i] = ( vBuffer[i] - fMinZ ) * fRange;
		}
	}
}