// Ryan Schmidt - rms@unknownroad.com
// Copyright (c) 2004-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt


#ifndef __RMS_WMLEXT_POLYGON_H__
#define __RMS_WMLEXT_POLYGON_H__

#include "WmlExt.h"

#include <WmlVector2.h>
#include <WmlMatrix2.h>
#include <WmlAxisAlignedBox2.h>
#include "WmlExtVectorUtil.h"

#include <vector>

namespace Wml {

template<class Real>
class WMLEXT_ITEM Polygon2
{
public:

	Polygon2();
	Polygon2( const Polygon2<Real> & copy );
	~Polygon2();

	// add/remove vertex functions
	inline unsigned int AppendVertex( const Wml::Vector2<Real> & vVertex );
	inline void ClearVertices();

	// vertex count stuff
	inline unsigned int VertexCount() const;
	inline bool IsEmpty() const;

	// element access
	inline const Wml::Vector2<Real> & Vertex( unsigned int nVertex ) const;
	inline Wml::Vector2<Real> & Vertex( unsigned int nVertex );
	inline Wml::Vector2<Real> & operator[]( int i );
	inline const Wml::Vector2<Real> & operator[] (int i ) const;
	inline const std::vector< Wml::Vector2<Real> > & Vertices() const;

	// orientation bits
	bool IsClockwise() const;
	bool IsCCW() const { return ! IsClockwise(); }
	void FlipOrientation();

	// inside/outside test
	bool IsInside( const Wml::Vector2<Real> & vTest ) const;

	// bounding things
	void BoundingBox( Wml::AxisAlignedBox2<Real> & dest ) const;

	// translation / rotation functions
	void Transform( const Wml::Matrix2<Real> & transform );
	void Translate( Real fX, Real fY );

	// simplify polygon
	void Simplify( bool bSimplifyStraightLines = true );

	// get distance to polygon
	Real SqrDistance( const Wml::Vector2<Real> & vPoint, Wml::Vector2<Real> * pNearest = NULL );

	// save and load to file
	bool Load( const char * pFilename );
	bool Save( const char * pFilename ) const;

	//! returns false if polygon is not counter-clockwise. pValues must be big enough to hold VertexCount() elements
	enum BarycentricType {
		Wachspress, DesbrunAngle, MeanValue };
	bool BarycentricCoords( const Wml::Vector2<Real> & vPoint, Real * pValues, BarycentricType eType = MeanValue ) const;

protected:
	std::vector< Wml::Vector2<Real> > m_vVertices;
};

typedef Polygon2<float> Polygon2f;
typedef Polygon2<double> Polygon2d;


/*
 * inlines
 */

template<class Real>
inline unsigned int Polygon2<Real>::AppendVertex( const Wml::Vector2<Real> & vVertex )
{
	size_t nVerts = m_vVertices.size();
	m_vVertices.push_back(vVertex);
	return (unsigned int)nVerts;
}

template<class Real>
inline void Polygon2<Real>::ClearVertices( )
{
	m_vVertices.resize(0);
}

template<class Real>
inline unsigned int Polygon2<Real>::VertexCount() const
{
	return (unsigned int)m_vVertices.size();
}

template<class Real>
inline bool Polygon2<Real>::IsEmpty() const
{
	return m_vVertices.empty();
}

template<class Real>
inline const Wml::Vector2<Real> & Polygon2<Real>::Vertex( unsigned int nVertex ) const
{
	return m_vVertices[nVertex];
}

template<class Real>
inline Wml::Vector2<Real> & Polygon2<Real>::Vertex( unsigned int nVertex )
{
	return m_vVertices[nVertex];
}

template<class Real>
inline const Wml::Vector2<Real> & Polygon2<Real>::operator[]( int i ) const
{
	return m_vVertices[i];
}

template<class Real>
inline Wml::Vector2<Real> & Polygon2<Real>::operator[]( int i )
{
	return m_vVertices[i];
}

template<class Real>
inline const std::vector< Wml::Vector2<Real> > & Polygon2<Real>::Vertices() const
{
	return m_vVertices;
}




} // namespace rmssketch



#endif // __RMS_POLYGON_H__